package ch.psi.daq.queryrest.controller;

import java.io.IOException;
import java.util.Collection;
import java.util.List;
import java.util.Map.Entry;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.Resource;
import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.http.MediaType;
import org.springframework.validation.Validator;
import org.springframework.web.bind.WebDataBinder;
import org.springframework.web.bind.annotation.InitBinder;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.RestController;

import ch.psi.daq.cassandra.request.validate.RequestProviderValidator;
import ch.psi.daq.common.concurrent.singleton.Deferred;
import ch.psi.daq.common.json.deserialize.AttributeBasedDeserializer;
import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.query.analyzer.QueryAnalyzer;
import ch.psi.daq.query.config.QueryConfig;
import ch.psi.daq.query.model.Aggregation;
import ch.psi.daq.query.model.AggregationType;
import ch.psi.daq.query.model.DBMode;
import ch.psi.daq.query.model.Query;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.query.model.ResponseFormat;
import ch.psi.daq.query.model.impl.DAQQuery;
import ch.psi.daq.query.processor.QueryProcessor;
import ch.psi.daq.query.request.ChannelsRequest;
import ch.psi.daq.queryrest.response.csv.CSVResponseStreamWriter;
import ch.psi.daq.queryrest.response.json.JSONResponseStreamWriter;

import com.google.common.collect.Lists;

@RestController
public class QueryRestController {

   private static final Logger LOGGER = LoggerFactory.getLogger(QueryRestController.class);

   public static final String CHANNELS = "/channels";
   public static final String QUERY = "/query";

   @Resource
   private Validator queryValidator;
   private Validator requestProviderValidator = new RequestProviderValidator();

   @Resource
   private JSONResponseStreamWriter jsonResponseStreamWriter;

   @Resource
   private CSVResponseStreamWriter csvResponseStreamWriter;

   @Resource
   private ApplicationContext appContext;

   // using Deferred ensures that data-api startup succeeds even if DataBuffer/ArchiverAppliance is
   // not reachable at startup
   private Deferred<QueryProcessor> cassandraQueryProcessor = new Deferred<QueryProcessor>(
         () -> appContext.getBean(QueryConfig.BEAN_NAME_CASSANDRA_QUERY_PROCESSOR, QueryProcessor.class));
   private Deferred<QueryProcessor> archiverApplianceQueryProcessor = new Deferred<QueryProcessor>(
         () -> appContext.getBean(QueryConfig.BEAN_NAME_ARCHIVER_APPLIANCE_QUERY_PROCESSOR, QueryProcessor.class));

   @Resource
   private Function<Query, QueryAnalyzer> queryAnalizerFactory;

   @InitBinder
   protected void initBinder(WebDataBinder binder) {
      if (binder.getTarget() != null) {
         if (requestProviderValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(requestProviderValidator);
         }
         if (queryValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(queryValidator);
         }
      }
   }

   @RequestMapping(value = CHANNELS, method = {RequestMethod.GET, RequestMethod.POST},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody Collection<String> getChannels(@RequestBody(required = false) ChannelsRequest request)
         throws Throwable {
      // in case not specified use default (e.g. GET)
      if (request == null) {
         request = new ChannelsRequest();
      }

      try {
         // sorted collection
         Collection<String> allChannels = getQueryProcessor(request.getDbMode()).getChannels(request.getRegex());
         return allChannels;
      } catch (Throwable t) {
         LOGGER.error("Failed to query channel names.", t);
         throw t;
      }
   }

   /**
    * Query specific channel names, and return only those.
    * 
    * @param channelName part of (or full) channel name
    * @return Collection of channel names matching the specified input channel name
    * @throws Throwable in case something goes wrong
    */
   @RequestMapping(value = CHANNELS + "/{channelName}", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody Collection<String> getChannels(@PathVariable(value = "channelName") String channelName)
         throws Throwable {
      return getChannels(new ChannelsRequest(channelName));
   }


   /**
    * Catch-all query method for getting data from the backend for both JSON and CSV requests.
    * <p>
    * The {@link DAQQuery} object will be a concrete subclass based on the combination of fields
    * defined in the user's query. The {@link AttributeBasedDeserializer} decides which class to
    * deserialize the information into and has been configured (see
    * QueryRestConfig#afterPropertiesSet) accordingly.
    * 
    * @param query concrete implementation of {@link DAQQuery}
    * @param res the {@link HttpServletResponse} instance associated with this request
    * @throws IOException thrown if writing to the output stream fails
    */
   @RequestMapping(
         value = QUERY,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public void executeQuery(@RequestBody @Valid DAQQuery query, HttpServletResponse res) throws Exception {
      try {
         LOGGER.debug("Executing query '{}'", query.toString());

         if (ResponseFormat.JSON.equals(query.getResponseFormat())) {
            // write the response back to the client using java 8 streams
            jsonResponseStreamWriter.respond(executeQuery(query), query, res);
         } else {
            // it's a CSV request
            executeQueryCsv(query, res);
         }
         
      } catch (Exception e) {
         LOGGER.error("Failed to execute query '{}'.", query, e);
         throw e;
      }
   }

   /**
    * Returns data in CSV format to the client.
    */
   private void executeQueryCsv(DAQQuery query, HttpServletResponse res) throws Exception {
      
      if (!(query.getAggregationType() == null || AggregationType.value.equals(query.getAggregationType()))) {
         // We allow only no aggregation or value aggregation as
         // extrema: nested structure and not clear how to map it to one line
         // index: value is an array of Statistics whose size is not clear at initialization time
         String message = "CSV export does not support '" + query.getAggregationType() + "'";
         LOGGER.warn(message);
         throw new IllegalArgumentException(message);
      }
      
      try {
         LOGGER.debug("Executing query '{}'", query.toString());
         // write the response back to the client using java 8 streams
         csvResponseStreamWriter.respond(executeQuery(query), query, res);
      } catch (Exception e) {
         LOGGER.error("Failed to execute query '{}'.", query, e);
         throw e;
      }
      
   }

   public Stream<Entry<String, ?>> executeQuery(DAQQuery query) {
      QueryAnalyzer queryAnalizer = queryAnalizerFactory.apply(query);

      // all the magic happens here
      Stream<Entry<String, Stream<? extends DataEvent>>> channelToDataEvents =
            getQueryProcessor(query.getDbMode()).process(queryAnalizer);

      // do post-process
      Stream<Entry<String, ?>> channelToData = queryAnalizer.postProcess(channelToDataEvents);

      return channelToData;
   }

   private QueryProcessor getQueryProcessor(DBMode dbMode) {
      if (DBMode.databuffer.equals(dbMode)) {
         return cassandraQueryProcessor.get();
      } else if (DBMode.archiverappliance.equals(dbMode)) {
         return archiverApplianceQueryProcessor.get();
      } else {
         LOGGER.error("Unknown DBMode '{}'!", dbMode);
         throw new IllegalArgumentException(String.format("Unknown DBMode '%s'", dbMode));
      }
   }

   /**
    * Returns the current list of {@link Ordering}s available.
    * 
    * @return list of {@link Ordering}s as String array
    */
   @RequestMapping(value = "ordering", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<String> getOrderingValues() {
      List<Ordering> orderings = Lists.newArrayList(Ordering.values());
      return orderings.stream()
            .map((Ordering ord) -> {
               return ord.toString();
            }).collect(Collectors.toList());
   }

   /**
    * Returns the current list of {@link QueryField}s available.
    * 
    * @return list of {@link QueryField}s as String array
    */
   @RequestMapping(value = "queryfields", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<String> getQueryFieldValues() {
      List<QueryField> orderings = Lists.newArrayList(QueryField.values());
      return orderings.stream()
            .map((QueryField qf) -> {
               return qf.toString();
            }).collect(Collectors.toList());
   }

   /**
    * Returns the current list of {@link Aggregation}s available.
    * 
    * @return list of {@link Aggregation}s as String array
    */
   @RequestMapping(value = "aggregations", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<String> getAggregationsValues() {
      List<Aggregation> orderings = Lists.newArrayList(Aggregation.values());
      return orderings.stream()
            .map((Aggregation value) -> {
               return value.toString();
            }).collect(Collectors.toList());
   }

   /**
    * Returns the current list of {@link AggregationType}s available.
    * 
    * @return list of {@link AggregationType}s as String array
    */
   @RequestMapping(value = "aggregationtypes", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<String> getAggregationTypesValues() {
      List<AggregationType> orderings = Lists.newArrayList(AggregationType.values());
      return orderings.stream()
            .map((AggregationType value) -> {
               return value.toString();
            }).collect(Collectors.toList());
   }

   /**
    * Returns the current list of {@link DBMode} values.
    * 
    * @return list of {@link DBMode}s as String array
    */
   @RequestMapping(value = "dbmodes", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<String> getDBModeValues() {
      List<DBMode> orderings = Lists.newArrayList(DBMode.values());
      return orderings.stream()
            .map((DBMode value) -> {
               return value.toString();
            }).collect(Collectors.toList());
   }
}
