package ch.psi.daq.queryrest.response.json;

import java.io.IOException;
import java.io.OutputStream;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.Stream;

import javax.annotation.Resource;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.MediaType;

import ch.psi.daq.query.model.Aggregation;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.query.model.impl.DAQQuery;
import ch.psi.daq.queryrest.response.AbstractResponseStreamWriter;

import com.fasterxml.jackson.core.JsonEncoding;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectWriter;
import com.fasterxml.jackson.databind.ser.impl.SimpleBeanPropertyFilter;
import com.fasterxml.jackson.databind.ser.impl.SimpleFilterProvider;

/**
 * Takes a Java 8 stream and writes it to the output stream provided by the {@link ServletResponse}
 * of the current request.
 */
public class JSONResponseStreamWriter extends AbstractResponseStreamWriter {

   private static final String DATA_RESP_FIELD = "data";

   private static final Logger logger = LoggerFactory.getLogger(JSONResponseStreamWriter.class);

   @Resource
   private JsonFactory jsonFactory;

   @Resource
   private ObjectMapper mapper;

   @Override
   public void respond(Stream<Entry<String, ?>> stream, DAQQuery query, HttpServletResponse response) throws Exception {
      response.setCharacterEncoding(JsonEncoding.UTF8.getJavaName());
      response.setContentType(MediaType.APPLICATION_JSON_VALUE);

      Set<String> includedFields = getFields(query);

      ObjectWriter writer = configureWriter(includedFields);
      
      respondInternal(stream, response, writer, query);
   }

   protected Set<String> getFields(DAQQuery query) {
      Set<QueryField> queryFields = query.getFields();
      List<Aggregation> aggregations = query.getAggregations();

      Set<String> includedFields =
            new LinkedHashSet<String>(queryFields.size() + (aggregations != null ? aggregations.size() : 0));

      for (QueryField field : queryFields) {
         includedFields.add(field.name());
      }
      if (aggregations != null) {
         for (Aggregation aggregation : query.getAggregations()) {
            includedFields.add(aggregation.name());
         }
      }

      // do not write channel since it is already provided as key in mapping
      includedFields.remove(QueryField.channel.name());

      return includedFields;
   }

   /**
    * Configures the writer dynamically by including the fields which should be included in the
    * response.
    * 
    * @param includedFields set of strings which correspond to the getter method names of the
    *        classes registered as a mixed-in
    * @return the configured writer that includes the specified fields
    */
   private ObjectWriter configureWriter(Set<String> includedFields) {
      SimpleFilterProvider propertyFilter = new SimpleFilterProvider();
      propertyFilter.addFilter("namedPropertyFilter", SimpleBeanPropertyFilter.filterOutAllExcept(includedFields));
      // only write the properties not excluded in the filter
      ObjectWriter writer = mapper.writer(propertyFilter);
      return writer;
   }

   /**
    * Writes the outer Java stream into the output stream.
    * 
    * @param stream Mapping from channel name to data
    * @param response {@link ServletResponse} instance given by the current HTTP request
    * @param writer configured writer that includes the fields the end user wants to see
    * @param query 
    * @throws IOException thrown if writing to the output stream fails
    */
   private void respondInternal(Stream<Entry<String, ?>> stream, HttpServletResponse response, ObjectWriter writer, DAQQuery query)
         throws Exception {
      
      OutputStream out = handleCompressionAndResponseHeaders(query, response, CONTENT_TYPE_JSON);
      JsonGenerator generator = jsonFactory.createGenerator(out, JsonEncoding.UTF8);

      AtomicReference<Exception> exception = new AtomicReference<>();
      try {
         generator.writeStartArray();
         stream
               /* ensure elements are sequentially written */
               .sequential()
               .forEach(
                     entry -> {
                        try {
                           generator.writeStartObject();
                           generator.writeStringField(QueryField.channel.name(), entry.getKey());

                           generator.writeFieldName(DATA_RESP_FIELD);
                           writer.writeValue(generator, entry.getValue());

                           generator.writeEndObject();
                        } catch (Exception e) {
                           logger.error("Could not write channel name of channel '{}'", entry.getKey(), e);
                           exception.compareAndSet(null, e);
                        }
                     });
         generator.writeEndArray();
      } finally {
         generator.flush();
         generator.close();
      }

      if (exception.get() != null) {
         throw exception.get();
      }
   }

}
