package ch.psi.daq.test.queryrest.controller;

import javax.annotation.Resource;

import org.junit.After;
import org.junit.Test;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.result.MockMvcResultHandlers;
import org.springframework.test.web.servlet.result.MockMvcResultMatchers;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import ch.psi.daq.cassandra.request.range.RequestRangeDate;
import ch.psi.daq.cassandra.request.range.RequestRangePulseId;
import ch.psi.daq.cassandra.request.range.RequestRangeTime;
import ch.psi.daq.cassandra.util.test.CassandraDataGen;
import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.query.model.AggregationType;
import ch.psi.daq.query.model.Compression;
import ch.psi.daq.query.model.impl.DAQQuery;
import ch.psi.daq.queryrest.controller.QueryRestController;
import ch.psi.daq.queryrest.filter.CorsFilter;
import ch.psi.daq.test.cassandra.admin.CassandraTestAdmin;
import ch.psi.daq.test.queryrest.AbstractDaqRestTest;

/**
 * Tests the {@link DaqController} implementation.
 */
public class QueryRestControllerJsonTest extends AbstractDaqRestTest {

   @Resource
   private CassandraTestAdmin cassandraTestAdmin;

   @Resource
   private CassandraDataGen dataGen;

   @Resource
   private CorsFilter corsFilter;

   public static final String TEST_CHANNEL_01 = "testChannel1";
   public static final String TEST_CHANNEL_02 = "testChannel2";
   public static final String[] TEST_CHANNEL_NAMES = new String[] {TEST_CHANNEL_01, TEST_CHANNEL_02};

   @After
   public void tearDown() throws Exception {}

   @Test
   public void testChannelNameQuery() throws Exception {

      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .get(QueryRestController.CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").value("BoolScalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").value("BoolWaveform"));

   }

   @Test
   public void testSpecificChannelSearch() throws Exception {
      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .get(QueryRestController.CHANNELS + "/int32")
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").value("Int32Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").value("Int32Waveform"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2]").value("UInt32Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[3]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[3]").value("UInt32Waveform"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[4]").doesNotExist());
   }

   @Test
   public void testCorsFilterNoHeaders() throws Exception {
      corsFilter.setForceAllHeaders(false);
      this.mockMvc = MockMvcBuilders.webAppContextSetup(webApplicationContext).addFilters(corsFilter).build();

      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .options(QueryRestController.CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            // we didn't set the 'Origin' header so no access-control
            .andExpect(MockMvcResultMatchers.header().doesNotExist("Access-Control-Allow-Origin"));
   }

   @Test
   public void testCorsFilterIncludesHeaders() throws Exception {
      // all headers are set
      this.mockMvc = MockMvcBuilders.webAppContextSetup(webApplicationContext).addFilters(corsFilter).build();

      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .options(QueryRestController.CHANNELS)
                  .header("Origin", "*")
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            // we didn't set the 'Origin' header so no access-control
            .andExpect(MockMvcResultMatchers.header().string("Access-Control-Allow-Origin", "*"));

      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .options(QueryRestController.CHANNELS)
                  .header("Origin", "http://localhost:8080")
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            // we didn't set the 'Origin' header so no access-control
            .andExpect(MockMvcResultMatchers.header().string("Access-Control-Allow-Origin", "http://localhost:8080"));

      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .options(QueryRestController.CHANNELS)
                  .header("Origin", "someBogusDomain.com")
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Access-Control-Allow-Origin", "*"));

   }

   @Test
   public void testCorsFilterMismatchSpecificOrigin() throws Exception {
      corsFilter.setForceAllHeaders(false);
      this.mockMvc = MockMvcBuilders.webAppContextSetup(webApplicationContext).addFilters(corsFilter).build();

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .options(QueryRestController.CHANNELS)
                        .header("Origin", "*")
                        .header("Access-Control-Request-Method", "GET")
                        .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Access-Control-Allow-Origin", "*"))
            .andExpect(
                  MockMvcResultMatchers.header().string("Access-Control-Allow-Headers",
                        "Origin, Authorization, Accept, Content-Type"));

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .options(QueryRestController.CHANNELS)
                        .header("Origin", "someBogusDomain.com")
                        .header("Access-Control-Request-Method", "GET")
                        .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().doesNotExist("Access-Control-Allow-Origin"))
            .andExpect(
                  MockMvcResultMatchers.header().string("Access-Control-Allow-Headers",
                        "Origin, Authorization, Accept, Content-Type"));
   }


   @Test
   public void testPulseRangeQuery() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setCompression(Compression.NONE);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalMillis").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(11))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalMillis").value(110))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalMillis").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(11))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalMillis").value(110));
   }

   @Test
   public void testTimeRangeQuery() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  100,
                  110),
            TEST_CHANNEL_NAMES);
      request.setCompression(Compression.NONE);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(QueryRestController.QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalMillis").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(11))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalMillis").value(110))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalMillis").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(11))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalMillis").value(110));
   }

   @Test
   public void testDateRangeQuery() throws Exception {
      String startDate = RequestRangeDate.format(100);
      String endDate = RequestRangeDate.format(110);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            TEST_CHANNEL_NAMES);
      request.setCompression(Compression.NONE);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(QueryRestController.QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content)
            )
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalMillis").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(11))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalMillis").value(110))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalMillis").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(11))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalMillis").value(110));
   }

   @Test
   public void testExtremaAggregation() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            false,
            Ordering.asc,
            AggregationType.extrema,
            TEST_CHANNEL_NAMES[0]);
      request.setCompression(Compression.NONE);

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders.post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())

            .andExpect(MockMvcResultMatchers.jsonPath("$").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").value(TEST_CHANNEL_NAMES[0]))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.value").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.value").value(Double.valueOf(100)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.event").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.event.pulseId").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.event.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.value").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.value").value(Double.valueOf(101)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.event").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.event.pulseId").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.event.pulseId").value(101));
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregate() throws Exception {
      long startTime = 100;
      long endTime = 199;
      String startDate = RequestRangeDate.format(startTime);
      String endDate = RequestRangeDate.format(endTime);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            TEST_CHANNEL_01);
      request.setNrOfBins(2);
      request.setCompression(Compression.NONE);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(QueryRestController.QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content)
            )
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalMillis").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(15))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalMillis").value(150))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(5));
   }

   @Test
   public void testDateRangeQueryBinSizeAggregate() throws Exception {
       long startTime = 1000;
       long endTime = 1999;
       String startDate = RequestRangeDate.format(startTime);
       String endDate = RequestRangeDate.format(endTime);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            TEST_CHANNEL_01);
      request.setBinSize(100);
      request.setCompression(Compression.NONE);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(QueryRestController.QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content)
            )
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(110))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalMillis").value(1100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].pulseId").value(120))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].globalMillis").value(1200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].pulseId").value(130))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].globalMillis").value(1300))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].pulseId").value(140))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].globalMillis").value(1400))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].pulseId").value(150))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].globalMillis").value(1500))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].pulseId").value(160))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].globalMillis").value(1600))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].pulseId").value(170))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].globalMillis").value(1700))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].pulseId").value(180))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].globalMillis").value(1800))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].pulseId").value(190))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].globalMillis").value(1900))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].eventCount").value(10));
   }
   
   @Test
   public void testGzipCompression() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setCompression(Compression.GZIP);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk());
   }
   
   @Test
   public void testGzipFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setCompression(Compression.GZIP);

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.gz"));
   }
   
   @Test
   public void testJsonFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setCompression(Compression.NONE);

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.json"));
   }
   
}
