/**
 * 
 */
package ch.psi.daq.test.queryrest.query;

import java.util.List;
import java.util.Random;
import java.util.concurrent.CompletableFuture;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.LongStream;
import java.util.stream.Stream;

import com.google.common.collect.Lists;

import ch.psi.daq.cassandra.reader.CassandraReader;
import ch.psi.daq.cassandra.reader.query.PulseIdRangeQuery;
import ch.psi.daq.cassandra.reader.query.TimeRangeQuery;
import ch.psi.daq.cassandra.util.test.CassandraDataGen;
import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.cassandra.ChannelEvent;
import ch.psi.daq.domain.cassandra.MetaPulseId;
import ch.psi.daq.domain.cassandra.MetaTime;
import ch.psi.daq.domain.cassandra.querying.ChannelEventQuery;
import ch.psi.daq.domain.cassandra.querying.ChannelEventQueryInfo;
import ch.psi.daq.domain.cassandra.querying.EventQuery;
import ch.psi.data.converters.ConverterProvider;

/**
 * @author zellweger_c
 *
 */
public class DummyCassandraReader implements CassandraReader {

   private static final int KEYSPACE = 1;
   private CassandraDataGen dataGen;
   private String[] channels;
   private Random random = new Random(0);

   /**
    * 
    */
   public DummyCassandraReader() {
      this.dataGen = new CassandraDataGen();
      this.channels = new String[] {
            "testChannel1",
            "testChannel2",
            "BoolScalar",
            "BoolWaveform",
            "Int8Scalar",
            "Int8Waveform",
            "UInt8Scalar",
            "UInt8Waveform",
            "Int16Scalar",
            "Int16Waveform",
            "UInt16Scalar",
            "UInt16Waveform",
            "Int32Scalar",
            "Int32Waveform",
            "UInt32Scalar",
            "UInt32Waveform",
            "Int64Scalar",
            "Int64Waveform",
            "UInt64Scalar",
            "UInt64Waveform",
            "Float32Scalar",
            "Float32Waveform",
            "Float64Scalar",
            "Float64Waveform",
            "StringScalar"};
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<String> getChannelStream(String regex) {
      Stream<String> channelStream = Stream.of(channels);
      if (regex != null) {
         Pattern pattern = Pattern.compile(regex.toLowerCase());
         channelStream = Stream.of(channels).filter(channel -> pattern.matcher(channel.toLowerCase()).find());
      }
      return channelStream;
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<? extends DataEvent> getEventStream(String channel, long startPulseId, long endPulseId,
         Ordering ordering, String... columns) {
      return getDummyEventStream(channel, startPulseId, endPulseId);
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<? extends DataEvent> getEventStream(String channel, long startMillis, long startNanos, long endMillis,
         long endNanos, Ordering ordering, String... columns) {
      return getDummyEventStream(channel, startMillis / 10, endMillis / 10);
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<ChannelEvent> getEventStream(EventQuery eventQuery, Stream<? extends ChannelEventQuery> queryProviders) {
      Stream<ChannelEvent> result = queryProviders.map(ceq -> {
         if (ceq instanceof ChannelEventQueryInfo) {
            return getEvent((ChannelEventQueryInfo) ceq);
         } else {
            throw new UnsupportedOperationException("This is not yet implemented!");
         }
      });

      return result;
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<ChannelEvent> getEventStream(PulseIdRangeQuery query) {
      Stream<ChannelEvent> dummyEventStream =
            getDummyEventStream(query.getChannel(), query.getStartPulseId(), query.getEndPulseId())
                  .map(ce -> {
                     return (ChannelEvent) ce;
                  });
      return dummyEventStream;

   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<ChannelEvent> getEventStream(TimeRangeQuery query) {
      Stream<ChannelEvent> dummyEventStream =
            getDummyEventStream(query.getChannel(), query.getStartMillis() / 10, query.getEndMillis() / 10)
                  .map(ce -> {
                     return (ChannelEvent) ce;
                  });
      return dummyEventStream;
   }


   private Stream<? extends DataEvent> getDummyEventStream(String channel, long startIndex, long endIndex) {
      String channelLower = channel.toLowerCase();

      Stream<? extends DataEvent> eventStream = LongStream.rangeClosed(startIndex, endIndex).mapToObj(i -> {
         if (channelLower.contains("waveform")) {
            long[] value = random.longs(2048).toArray();
            value[0] = i;
            return new ChannelEvent(
                  channel,
                  i * 10,
                  0,
                  i,
                  i * 10,
                  0,
                  value
            );

         } else if (channelLower.contains("image")) {
            int x = 640;
            int y = 480;
            int[] shape = new int[] {x, y};
            long[] value = random.longs(x * y).toArray();
            value[0] = i;
            return new ChannelEvent(
                  channel,
                  i * 10,
                  0,
                  i,
                  i * 10,
                  0,
                  value,
                  shape
            );
         } else {
            return new ChannelEvent(
                  channel,
                  i * 10,
                  0,
                  i,
                  i * 10,
                  0,
                  i
            );
         }
      });

      return eventStream;
   }

   private List<? extends DataEvent> getDummyEvents(String channel, long startIndex, long endIndex) {
      return getDummyEventStream(channel, startIndex, endIndex).collect(Collectors.toList());
   }

   /**
    * @{inheritDoc
    */
   @Override
   public List<String> getChannels() {
      return Lists.newArrayList(channels);
   }

   /**
    * @{inheritDoc
    */
   @Override
   public List<String> getChannels(String regex) {
      return Lists.newArrayList(channels).stream().filter(s -> {
         return s.contains(regex);
      }).collect(Collectors.toList());
   }

   /**
    * @{inheritDoc
    */
   @Override
   public ChannelEvent getEvent(ChannelEventQueryInfo queryInfo, String... columns) {
      if (queryInfo.getPulseId() > 0) {
         return (ChannelEvent) getDummyEvents(queryInfo.getChannel(), queryInfo.getPulseId(), queryInfo.getPulseId())
               .get(0);
      }
      return (ChannelEvent) getDummyEvents(queryInfo.getChannel(), queryInfo.getGlobalMillis() / 10,
            queryInfo.getGlobalMillis() / 10).get(0);
   }

   /**
    * @{inheritDoc
    */
   @Override
   public CompletableFuture<ChannelEvent> getEventAsync(ChannelEventQueryInfo queryInfo, String... columns) {
      throw new UnsupportedOperationException();
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<? extends ChannelEventQuery> getChannelEventQueryStream(PulseIdRangeQuery query) {

      return dataGen.generateMetaPulseId(
            KEYSPACE,
            ConverterProvider.TYPE_INT32,
            null,
            query.getStartPulseId(),
            (query.getEndPulseId() - query.getStartPulseId() + 1),
            i -> i * 10,
            i -> 0,
            i -> i,
            query.getChannel()).stream();
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<? extends ChannelEventQuery> getChannelEventQueryStream(TimeRangeQuery query) {

      return dataGen.generateMetaTime(
            KEYSPACE,
            ConverterProvider.TYPE_INT32,
            null,
            query.getStartMillis() / 10,
            ((query.getEndMillis() - query.getStartMillis()) / 10 + 1),
            i -> i * 10,
            i -> 0,
            i -> i,
            query.getChannel()).stream();
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<MetaPulseId> getMetaPulseIdStream(PulseIdRangeQuery query) {

      return getChannelEventQueryStream(query).map(r -> {
         return (MetaPulseId) r;
      });

   }


   /**
    * @{inheritDoc
    */
   @Override
   public Stream<MetaTime> getMetaTimeStream(TimeRangeQuery query) {

      return getChannelEventQueryStream(query).map(r -> {
         return (MetaTime) r;
      });
   }

}
